@@ -10,15 +10,13 @@
 
 import time
 from calendar import timegm
-from six import PY3
 import re
 
 import os
-import six.moves.builtins as __builtin__
+import builtins
 from os.path import join as opj
 
 from datalad.downloaders.tests.utils import get_test_providers
-from ..base import AccessFailedError
 from ..base import DownloadError
 from ..base import IncompleteDownloadError
 from ..base import BaseDownloader
@@ -30,26 +28,27 @@ from ..http import HTMLFormAuthenticator
 from ..http import HTTPDownloader
 from ..http import HTTPBearerTokenAuthenticator
 from ..http import process_www_authenticate
+from ...support.exceptions import AccessFailedError
 from ...support.network import get_url_straight_filename
 from ...tests.utils import with_fake_cookies_db
 from ...tests.utils import skip_if_no_network
 from ...tests.utils import with_testsui
 from ...tests.utils import with_memory_keyring
+from ...tests.utils import known_failure_githubci_win
 
 # BTW -- mock_open is not in mock on wheezy (Debian 7.x)
 try:
-    if PY3:
-        raise ImportError("Not yet ready apparently: https://travis-ci.org/datalad/datalad/jobs/111659666")
+    raise ImportError("Not yet ready apparently: https://travis-ci.org/datalad/datalad/jobs/111659666")
     import httpretty
 except (ImportError, AttributeError):
     # Attribute Error happens with newer httpretty and older ssl module
     # https://github.com/datalad/datalad/pull/2623
     class NoHTTPPretty(object):
-       __bool__ = __nonzero__ = lambda s: False
+       __bool__ = lambda s: False
        activate = lambda s, t: t
     httpretty = NoHTTPPretty()
 
-from mock import patch
+from unittest.mock import patch
 from ...tests.utils import SkipTest
 from ...tests.utils import assert_in
 from ...tests.utils import assert_not_in
@@ -107,6 +106,7 @@ def test_process_www_authenticate():
                  [])
 
 
+@known_failure_githubci_win
 @with_tree(tree=[('file.dat', 'abc')])
 @serve_path_via_http
 def test_HTTPDownloader_basic(toppath, topurl):
@@ -117,11 +117,16 @@ def test_HTTPDownloader_basic(toppath, topurl):
     download(furl, tfpath)
     ok_file_has_content(tfpath, 'abc')
 
-    # download() creates leading directories if needed.
+    # download() creates leading directories if needed for file targets...
     subdir_tfpath = opj(toppath, "l1", "l2", "file-downloaded.dat")
     download(furl, subdir_tfpath)
     ok_file_has_content(subdir_tfpath, 'abc')
 
+    # ... and for directory targets.
+    subdir_dirtarget = opj(toppath, "d1", "d2", "")
+    download(furl, subdir_dirtarget)
+    ok_file_has_content(opj(subdir_dirtarget, "file.dat"), "abc")
+
     # see if fetch works correctly
     assert_equal(downloader.fetch(furl), 'abc')
 
@@ -142,7 +147,7 @@ def test_HTTPDownloader_basic(toppath, topurl):
     # XXX obscure mocking since impossible to mock write alone
     # and it still results in some warning being spit out
     with swallow_logs(), \
-         patch.object(__builtin__, 'open', fake_open(write_=_raise_IOError)):
+         patch.object(builtins, 'open', fake_open(write_=_raise_IOError)):
         assert_raises(DownloadError, download, furl, tfpath, overwrite=True)
 
     # incomplete download scenario - should have 3 tries
@@ -313,6 +318,7 @@ def test_authenticate_external_portals():
 test_authenticate_external_portals.tags = ['external-portal', 'network']
 
 
+@known_failure_githubci_win
 @skip_if_no_network
 def test_download_ftp():
     try:
